import { jsx as _jsx } from "react/jsx-runtime";
import { isBoolean, isNumber, isUndefined, nth } from 'lodash';
import React, { useImperativeHandle, useMemo, useRef, useState } from 'react';
import { useEventCallback, useForkRef, useIsomorphicLayoutEffect, useRefExtra, useResize } from '@react-devui/hooks';
import { checkScrollEnd, toPx } from '@react-devui/utils';
import { cloneHTMLElement, registerComponentMate } from '../../utils';
import { useComponentConfig } from '../root';
const EMPTY = Symbol();
const { COMPONENT_NAME } = registerComponentMate({ COMPONENT_NAME: 'DVirtualScroll' });
function VirtualScroll(props, ref) {
    const { children, dRef, dList, dFillNode, dItemRender, dItemSize, dItemNested, dItemKey, dFocusable = true, dFocusItem, dSize, dPadding, dHorizontal = false, dEmptyRender, dExpands, onScrollEnd, } = useComponentConfig(COMPONENT_NAME, props);
    //#region Ref
    const listRef = useRefExtra(undefined, true);
    const combineListlRef = useForkRef(listRef, dRef === null || dRef === void 0 ? void 0 : dRef.list);
    //#endregion
    const dataRef = useRef({
        listCache: new Map(),
    });
    const [scrollPosition, setScrollPosition] = useState(0);
    const getItemSize = useMemo(() => (isNumber(dItemSize) ? () => dItemSize : dItemSize), [dItemSize]);
    const checkFocusable = useMemo(() => (isBoolean(dFocusable) ? () => dFocusable : dFocusable), [dFocusable]);
    const [itemsMap, totalSize, firstFocusableItem, lastFocusableItem] = useMemo(() => {
        let accSize = 0;
        let firstFocusableItem;
        let lastFocusableItem;
        const items = new Map();
        const reduceArr = (arr, level = 0) => {
            var _a;
            let size = 0;
            for (const item of arr) {
                if (checkFocusable(item)) {
                    lastFocusableItem = item;
                    if (isUndefined(firstFocusableItem)) {
                        firstFocusableItem = item;
                    }
                }
                const key = dItemKey(item);
                const itemSize = getItemSize(item);
                size += itemSize;
                accSize += itemSize;
                const data = { item, level, accSize, nestedSize: 0 };
                items.set(key, data);
                const nestedData = dItemNested === null || dItemNested === void 0 ? void 0 : dItemNested(item);
                if (nestedData && nestedData.list && (isUndefined(dExpands) || dExpands.has(key))) {
                    if (nestedData.list.length === 0) {
                        data.nestedSize = (_a = nestedData.emptySize) !== null && _a !== void 0 ? _a : 0;
                        size += data.nestedSize;
                        accSize += data.nestedSize;
                    }
                    else {
                        data.nestedSize = reduceArr(nestedData.list, level + 1);
                        size += data.nestedSize;
                    }
                }
            }
            return size;
        };
        return [items, reduceArr(dList), firstFocusableItem, lastFocusableItem];
    }, [checkFocusable, dExpands, dItemKey, dItemNested, dList, getItemSize]);
    const [elSize, setElSize] = useState();
    const [elPaddingSize, setElPaddingSize] = useState();
    const getSize = () => {
        if (listRef.current) {
            if (isUndefined(dSize)) {
                setElSize(listRef.current[dHorizontal ? 'clientWidth' : 'clientHeight']);
            }
            if (isUndefined(dPadding)) {
                setElPaddingSize(toPx(getComputedStyle(listRef.current).getPropertyValue(dHorizontal ? 'padding-left' : 'padding-top'), true));
            }
        }
    };
    // eslint-disable-next-line react-hooks/exhaustive-deps
    useIsomorphicLayoutEffect(() => {
        getSize();
    });
    useResize(listRef, getSize);
    const ulSize = dSize !== null && dSize !== void 0 ? dSize : elSize;
    const paddingSize = dPadding !== null && dPadding !== void 0 ? dPadding : elPaddingSize;
    const list = (() => {
        if (isUndefined(ulSize) || isUndefined(paddingSize)) {
            return [];
        }
        const maxScrollSize = Math.max(totalSize + paddingSize * 2 - ulSize, 0);
        const _scrollPosition = Math.min(scrollPosition, maxScrollSize);
        let totalAccSize = 0;
        const startSize = _scrollPosition - ulSize - paddingSize;
        const endSize = _scrollPosition + ulSize + ulSize - paddingSize;
        let hasStart = false;
        let hasEnd = false;
        const getList = (arr, parent = []) => {
            var _a, _b, _c, _d;
            const fillSize = [0, 0];
            const list = [];
            const setsize = arr.filter((item) => {
                if (item === EMPTY) {
                    return false;
                }
                else {
                    const nestedData = dItemNested === null || dItemNested === void 0 ? void 0 : dItemNested(item);
                    if (nestedData && nestedData.list) {
                        return nestedData.inAriaSetsize;
                    }
                }
                return true;
            }).length;
            for (const [index, item] of arr.entries()) {
                let key = '';
                let size = 0;
                let nestedList;
                let childrenSize = 0;
                let emptyNode;
                if (item === EMPTY) {
                    size = (_b = (_a = dItemNested === null || dItemNested === void 0 ? void 0 : dItemNested(parent[parent.length - 1])) === null || _a === void 0 ? void 0 : _a.emptySize) !== null && _b !== void 0 ? _b : 0;
                    emptyNode = _jsx(React.Fragment, { children: dEmptyRender === null || dEmptyRender === void 0 ? void 0 : dEmptyRender(parent[parent.length - 1]) }, "$$empty");
                }
                else {
                    key = dItemKey(item);
                    size = getItemSize(item);
                    nestedList = (_c = dItemNested === null || dItemNested === void 0 ? void 0 : dItemNested(item)) === null || _c === void 0 ? void 0 : _c.list;
                    if (nestedList) {
                        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                        childrenSize = itemsMap.get(key).nestedSize;
                    }
                }
                if (hasEnd) {
                    fillSize[1] += size + childrenSize;
                    continue;
                }
                totalAccSize += size;
                if (nestedList) {
                    if (totalAccSize + childrenSize > startSize) {
                        let childrenList = [];
                        if (isUndefined(dExpands)) {
                            childrenList = getList(nestedList.length === 0 ? [EMPTY] : nestedList, parent.concat([item]));
                        }
                        else {
                            childrenList = (_d = dataRef.current.listCache.get(key)) !== null && _d !== void 0 ? _d : [];
                            if (dExpands.has(key)) {
                                childrenList = getList(nestedList.length === 0 ? [EMPTY] : nestedList, parent.concat([item]));
                                dataRef.current.listCache.set(key, childrenList);
                            }
                        }
                        list.push(dItemRender(item, index, {
                            aria: { 'aria-level': parent.length + 1, 'aria-setsize': setsize, 'aria-posinset': index + 1 },
                            vsList: childrenList,
                        }, parent));
                    }
                    else {
                        totalAccSize += childrenSize;
                        fillSize[0] += size + childrenSize;
                    }
                }
                else if (!hasStart) {
                    if (totalAccSize > startSize) {
                        list.push(item === EMPTY
                            ? emptyNode
                            : dItemRender(item, index, {
                                aria: { 'aria-level': parent.length + 1, 'aria-setsize': setsize, 'aria-posinset': index + 1 },
                            }, parent));
                        hasStart = true;
                    }
                    else {
                        fillSize[0] += size;
                    }
                }
                else if (!hasEnd) {
                    if (totalAccSize > endSize) {
                        hasEnd = true;
                        fillSize[1] += size;
                    }
                    else {
                        list.push(item === EMPTY
                            ? emptyNode
                            : dItemRender(item, index, {
                                aria: { 'aria-level': parent.length + 1, 'aria-setsize': setsize, 'aria-posinset': index + 1 },
                            }, parent));
                    }
                }
            }
            if (fillSize[0] > 0) {
                list.unshift(React.cloneElement(dFillNode, {
                    key: '$$fill-size-0',
                    style: Object.assign(Object.assign({}, dFillNode.props.style), { overflow: 'hidden', [dHorizontal ? 'height' : 'width']: 0, [dHorizontal ? 'width' : 'height']: fillSize[0], flexShrink: 0 }),
                    'aria-hidden': true,
                }));
            }
            if (fillSize[1] > 0) {
                list.push(React.cloneElement(dFillNode, {
                    key: '$$fill-size-1',
                    style: Object.assign(Object.assign({}, dFillNode.props.style), { overflow: 'hidden', [dHorizontal ? 'height' : 'width']: 0, [dHorizontal ? 'width' : 'height']: fillSize[1], flexShrink: 0 }),
                    'aria-hidden': true,
                }));
            }
            return list;
        };
        return getList(dList);
    })();
    const scrollTo = (num) => {
        if (listRef.current) {
            listRef.current[dHorizontal ? 'scrollLeft' : 'scrollTop'] = num;
        }
    };
    const scrollToItem = useEventCallback((item) => {
        if (!isUndefined(paddingSize)) {
            const findItem = itemsMap.get(dItemKey(item));
            if (!isUndefined(findItem)) {
                scrollTo(findItem.accSize - getItemSize(findItem.item) + paddingSize);
            }
        }
    });
    const scrollToStart = useEventCallback(() => {
        scrollTo(0);
        return firstFocusableItem;
    });
    const scrollToEnd = useEventCallback(() => {
        if (listRef.current) {
            scrollTo(listRef.current[dHorizontal ? 'scrollWidth' : 'scrollHeight']);
        }
        return lastFocusableItem;
    });
    const scrollToStep = useEventCallback((step) => {
        if (!isUndefined(paddingSize)) {
            if (isUndefined(dFocusItem)) {
                return;
            }
            let findItem;
            let offsetSize;
            if (listRef.current) {
                let index = -1;
                let findIndex = -1;
                const accSizeList = [];
                const focusKey = dItemKey(dFocusItem);
                for (const iterator of itemsMap) {
                    index += 1;
                    if (iterator[0] === focusKey) {
                        findIndex = index;
                    }
                    accSizeList.push(iterator[1]);
                }
                if (findIndex !== -1) {
                    if (step === 1) {
                        for (let index = findIndex + 1, n = 0; n < accSizeList.length; index++, n++) {
                            const accSizeItem = nth(accSizeList, index % accSizeList.length);
                            if (accSizeItem && checkFocusable(accSizeItem.item)) {
                                findItem = accSizeItem.item;
                                offsetSize = [accSizeItem.accSize - getItemSize(findItem) + paddingSize, accSizeItem.accSize + paddingSize];
                                break;
                            }
                        }
                    }
                    else {
                        for (let index = findIndex - 1, n = 0; n < accSizeList.length; index--, n++) {
                            const accSizeItem = nth(accSizeList, index);
                            if (accSizeItem && checkFocusable(accSizeItem.item)) {
                                findItem = accSizeItem.item;
                                offsetSize = [accSizeItem.accSize - getItemSize(findItem) + paddingSize, accSizeItem.accSize + paddingSize];
                                break;
                            }
                        }
                    }
                }
                if (!isUndefined(offsetSize)) {
                    const listElScrollPosition = listRef.current[dHorizontal ? 'scrollLeft' : 'scrollTop'];
                    const listElClientSize = listRef.current[dHorizontal ? 'clientWidth' : 'clientHeight'];
                    if (listElScrollPosition > offsetSize[1]) {
                        scrollTo(offsetSize[0] - paddingSize);
                    }
                    else if (offsetSize[0] > listElScrollPosition + listElClientSize) {
                        scrollTo(offsetSize[1] - listElClientSize + paddingSize);
                    }
                    else {
                        if (step === 1) {
                            if (offsetSize[1] > listElScrollPosition + listElClientSize) {
                                scrollTo(offsetSize[1] - listElClientSize + paddingSize);
                            }
                        }
                        else {
                            if (listElScrollPosition > offsetSize[0]) {
                                scrollTo(offsetSize[0] - paddingSize);
                            }
                        }
                    }
                }
            }
            return findItem;
        }
    });
    const scrollToNested = useEventCallback(() => {
        if (!isUndefined(paddingSize)) {
            if (isUndefined(dFocusItem)) {
                return;
            }
            let findItem;
            let offsetSize;
            if (listRef.current) {
                let index = -1;
                let findIndex = -1;
                let level = 0;
                const accSizeList = [];
                const focusKey = dItemKey(dFocusItem);
                for (const iterator of itemsMap) {
                    index += 1;
                    if (iterator[0] === focusKey) {
                        findIndex = index;
                        level = iterator[1].level;
                    }
                    accSizeList.push(iterator[1]);
                }
                if (findIndex !== -1) {
                    for (let index = findIndex + 1; index < accSizeList.length; index++) {
                        const accSizeItem = accSizeList[index];
                        if (accSizeItem.level <= level) {
                            return;
                        }
                        if (checkFocusable(accSizeItem.item)) {
                            findItem = accSizeItem.item;
                            offsetSize = [accSizeItem.accSize - getItemSize(findItem) + paddingSize, accSizeItem.accSize + paddingSize];
                            break;
                        }
                    }
                }
                if (!isUndefined(offsetSize)) {
                    const listElScrollPosition = listRef.current[dHorizontal ? 'scrollLeft' : 'scrollTop'];
                    const listElClientSize = listRef.current[dHorizontal ? 'clientWidth' : 'clientHeight'];
                    if (listElScrollPosition > offsetSize[1]) {
                        scrollTo(offsetSize[0] - paddingSize);
                    }
                    else if (offsetSize[0] > listElScrollPosition + listElClientSize) {
                        scrollTo(offsetSize[1] - listElClientSize + paddingSize);
                    }
                    else {
                        if (offsetSize[1] > listElScrollPosition + listElClientSize) {
                            scrollTo(offsetSize[1] - listElClientSize + paddingSize);
                        }
                    }
                }
            }
            return findItem;
        }
    });
    useImperativeHandle(ref, () => ({
        scrollToItem,
        scrollToStep,
        scrollToNested,
        scrollToStart,
        scrollToEnd,
    }), [scrollToItem, scrollToStep, scrollToNested, scrollToStart, scrollToEnd]);
    return children({
        render: (el) => cloneHTMLElement(el, {
            ref: combineListlRef,
            onScroll: (e) => {
                var _a, _b;
                (_b = (_a = el.props).onScroll) === null || _b === void 0 ? void 0 : _b.call(_a, e);
                setScrollPosition(e.currentTarget[dHorizontal ? 'scrollLeft' : 'scrollTop']);
                if (checkScrollEnd(e.currentTarget)[dHorizontal ? 'x' : 'y']) {
                    onScrollEnd === null || onScrollEnd === void 0 ? void 0 : onScrollEnd();
                }
            },
        }),
        vsList: list,
    });
}
export const DVirtualScroll = React.forwardRef(VirtualScroll);
